import type { Firestore, WriteResult } from 'firebase-admin/firestore';
import type { Either } from 'fp-ts/Either';
import * as E from 'fp-ts/Either';
import type { ReaderTaskEither } from 'fp-ts/ReaderTaskEither';
import * as RTE from 'fp-ts/ReaderTaskEither';
import * as TE from 'fp-ts/TaskEither';
import { pipe } from 'fp-ts/function';
import * as t from 'io-ts';
import { blogCodec, blogIdCodec } from './blogs';
import { musicCodec, musicIdCodec } from './music';

export const codecs = {
  blogs: { idCodec: blogIdCodec, valueCodec: blogCodec },
  music: { idCodec: musicIdCodec, valueCodec: musicCodec },
};

export type CollectionKey = keyof typeof codecs;

export type IdOf<K extends CollectionKey> = t.TypeOf<(typeof codecs)[K]['idCodec']>;
export type ValueOf<K extends CollectionKey> = t.TypeOf<(typeof codecs)[K]['valueCodec']>;

export type Entity<K extends CollectionKey> = { id: IdOf<K> } & ValueOf<K>;

export type FirestoreError =
  | {
      _tag: 'DecodeError';
      error: t.Errors;
    }
  | {
      _tag: 'InsertError';
      error: unknown;
    }
  | {
      _tag: 'UpdateError';
      error: unknown;
    }
  | {
      _tag: 'DeleteError';
      error: unknown;
    }
  | {
      _tag: 'ReadError';
      error: unknown;
    };

export const DecodeError = (error: t.Errors): FirestoreError => ({
  _tag: 'DecodeError',
  error,
});

export const InsertError = (error: unknown): FirestoreError => ({
  _tag: 'InsertError',
  error,
});

export const UpdateError = (error: unknown): FirestoreError => ({
  _tag: 'UpdateError',
  error,
});

export const DeleteError = (error: unknown): FirestoreError => ({
  _tag: 'DeleteError',
  error,
});

export const ReadError = (error: unknown): FirestoreError => ({
  _tag: 'ReadError',
  error,
});

// type IdDecoder<K extends CollectionKey> = (i: unknown) => t.Validation<IdOf<K>>;

const decodeId =
  <K extends CollectionKey>(collection: K) =>
  (id: string): Either<FirestoreError, IdOf<K>> => {
    const r = codecs[collection].idCodec.decode(id);
    if (E.isLeft(r)) {
      return E.left(DecodeError(r.left));
    } else {
      return E.right(r.right);
    }
    // return pipe(
    //   id,
    //   codecs[collection].idCodec.decode as IdDecoder<K>,
    //   E.mapLeft(DecodeError),
    // );
  };

const decodeValue =
  <K extends CollectionKey>(collection: K) =>
  (value: unknown): Either<FirestoreError, ValueOf<K>> => {
    const r = codecs[collection].valueCodec.decode(value);
    if (E.isLeft(r)) {
      return E.left(DecodeError(r.left));
    } else {
      return E.right(r.right);
    }
  };

export const create =
  <K extends CollectionKey>(collectionKey: K) =>
  (value: ValueOf<K>): ReaderTaskEither<Firestore, FirestoreError, IdOf<K>> => {
    const r = pipe(
      RTE.ask<Firestore>(),
      RTE.chainIOK((firestore) => () => firestore.collection(collectionKey)),
      RTE.chainTaskEitherK((collection) => TE.tryCatch(() => collection.add(value), InsertError)),
      RTE.chainEitherK((doc) => decodeId(collectionKey)(doc.id))
    );

    return r;
  };

export const read =
  <K extends CollectionKey>(collectionKey: K) =>
  (id: IdOf<K>): ReaderTaskEither<Firestore, FirestoreError, Entity<K>> => {
    const r = pipe(
      RTE.ask<Firestore>(),
      RTE.chainIOK((firestore) => () => firestore.collection(collectionKey)),
      RTE.chainIOK((collection) => () => collection.doc(id)),
      RTE.chainTaskEitherK((doc) => TE.tryCatch(() => doc.get(), ReadError)),
      RTE.chainIOK((snap) => () => snap.data()),
      RTE.chainEitherK(decodeValue(collectionKey)),
      RTE.map((value) => ({ id, ...value }))
    );

    return r;
  };

export const update =
  <K extends CollectionKey>(collectionKey: K) =>
  (id: IdOf<K>) =>
  (value: ValueOf<K>): ReaderTaskEither<Firestore, FirestoreError, WriteResult> => {
    const r = pipe(
      RTE.ask<Firestore>(),
      RTE.chainIOK((firestore) => () => firestore.collection(collectionKey)),
      RTE.chainIOK((collection) => () => collection.doc(id)),
      RTE.chainTaskEitherK((doc) => TE.tryCatch(() => doc.update(value), UpdateError))
    );

    return r;
  };

export const remove =
  <K extends CollectionKey>(collectionKey: K) =>
  (id: IdOf<K>): ReaderTaskEither<Firestore, FirestoreError, WriteResult> => {
    const r = pipe(
      RTE.ask<Firestore>(),
      RTE.chainIOK((firestore) => () => firestore.collection(collectionKey)),
      RTE.chainIOK((collection) => () => collection.doc(id)),
      RTE.chainTaskEitherK((doc) => TE.tryCatch(() => doc.delete(), DeleteError))
    );

    return r;
  };
